@@ -1,27 +1,32 @@
 #!/usr/bin/env python
-# test_ftpd.py
+# $Id: test_ftpd.py 619 2009-08-02 23:32:20Z g.rodola $
 
 #  ======================================================================
-#  Copyright (C) 2007 Giampaolo Rodola' <g.rodola@gmail.com>
+#  Copyright (C) 2007-2009 Giampaolo Rodola' <g.rodola@gmail.com>
 #
 #                         All Rights Reserved
 #
-#  Permission to use, copy, modify, and distribute this software and
-#  its documentation for any purpose and without fee is hereby
-#  granted, provided that the above copyright notice appear in all
-#  copies and that both that copyright notice and this permission
-#  notice appear in supporting documentation, and that the name of
-#  Giampaolo Rodola' not be used in advertising or publicity pertaining to
-#  distribution of the software without specific, written prior
-#  permission.
+# Permission is hereby granted, free of charge, to any person
+# obtaining a copy of this software and associated documentation
+# files (the "Software"), to deal in the Software without
+# restriction, including without limitation the rights to use,
+# copy, modify, merge, publish, distribute, sublicense, and/or sell
+# copies of the Software, and to permit persons to whom the
+# Software is furnished to do so, subject to the following
+# conditions:
+#
+# The above copyright notice and this permission notice shall be
+# included in all copies or substantial portions of the Software.
+#
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
+# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
+# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
+# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
+# OTHER DEALINGS IN THE SOFTWARE.
 #
-#  Giampaolo Rodola' DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
-#  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
-#  NO EVENT Giampaolo Rodola' BE LIABLE FOR ANY SPECIAL, INDIRECT OR
-#  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
-#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
-#  NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
-#  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 #  ======================================================================
 
 
@@ -47,6 +52,7 @@ import threading
 import unittest
 import socket
 import os
+import shutil
 import time
 import re
 import tempfile
@@ -534,6 +540,11 @@ class TestFtpAuthentication(unittest.TestCase):
         self.assertRaises(ftplib.error_perm, self.client.login, 'wrong', PASSWD)
         self.assertRaises(ftplib.error_perm, self.client.login, 'wrong', 'wrong')
 
+    def test_wrong_cmds_order(self):
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'pass ' + PASSWD)
+        self.client.login(user=USER, passwd=PASSWD)
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'pass ' + PASSWD)
+
     def test_max_auth(self):
         self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
         self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
@@ -654,17 +665,21 @@ class TestFtpDummyCmds(unittest.TestCase):
         self.client.sendcmd('type a')
         self.client.sendcmd('type i')
         self.client.sendcmd('type l7')
-        self.client.sendcmd('type l8')
+        self.client.sendcmd('type l8')       
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'type ?!?')
 
     def test_stru(self):
         self.client.sendcmd('stru f')
         self.client.sendcmd('stru F')
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stru p')
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stru r')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stru ?!?')
 
     def test_mode(self):
         self.client.sendcmd('mode s')
         self.client.sendcmd('mode S')
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'mode b')
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'mode c')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'mode ?!?')
 
     def test_noop(self):
@@ -814,7 +829,7 @@ class TestFtpFsOperations(unittest.TestCase):
         if os.path.exists(self.tempfile):
             os.remove(self.tempfile)
         if os.path.exists(self.tempdir):
-            os.rmdir(self.tempdir)
+            shutil.rmtree(self.tempdir)
 
     def test_cwd(self):
         self.client.cwd(self.tempdir)
@@ -831,10 +846,17 @@ class TestFtpFsOperations(unittest.TestCase):
         self.assertEqual(self.client.pwd(), '/' + self.tempdir)
 
     def test_cdup(self):
-        self.client.cwd(self.tempdir)
-        self.assertEqual(self.client.pwd(), '/' + self.tempdir)
+        subfolder = os.path.basename(tempfile.mkdtemp(dir=self.tempdir))
+        self.assertEqual(self.client.pwd(), '/')
+        self.client.cwd(self.tempdir)        
+        self.assertEqual(self.client.pwd(), '/%s' %self.tempdir)        
+        self.client.cwd(subfolder)
+        self.assertEqual(self.client.pwd(), '/%s/%s' %(self.tempdir, subfolder))
+        self.client.sendcmd('cdup')
+        self.assertEqual(self.client.pwd(), '/%s' %self.tempdir)
         self.client.sendcmd('cdup')
         self.assertEqual(self.client.pwd(), '/')
+
         # make sure we can't escape from root directory
         self.client.sendcmd('cdup')
         self.assertEqual(self.client.pwd(), '/')
@@ -878,6 +900,9 @@ class TestFtpFsOperations(unittest.TestCase):
         bogus = os.path.basename(tempfile.mktemp(dir=HOME))
         self.assertRaises(ftplib.error_perm, self.client.rename, bogus, '/x')
         self.assertRaises(ftplib.error_perm, self.client.rename, self.tempfile, '/')
+        # rnto sent without first specifying the source
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'rnto ' + self.tempfile)
+
         # make sure we can't rename root directory
         self.assertRaises(ftplib.error_perm, self.client.rename, '/', '/x')
 
@@ -891,9 +916,9 @@ class TestFtpFsOperations(unittest.TestCase):
         except ftplib.error_perm, err:
             self.failUnless("not retrievable" in str(err))
         else:
-            self.fail('Exception not raised')
+            self.fail('Exception not raised')      
 
-    def test_unforseen_mdtm_event(self):
+    def test_unforeseen_mdtm_event(self):
         # Emulate a case where the file last modification time is prior
         # to year 1900.  This most likely will never happen unless
         # someone specifically force the last modification time of a
@@ -907,7 +932,7 @@ class TestFtpFsOperations(unittest.TestCase):
         # ValueError as expected;
         # (see http://bugs.python.org/issue874042)
         try:
-            time.localtime(-9000000000)
+            time.strftime("%Y%m%d%H%M%S", time.localtime(-9000000000))
         except ValueError:
             skip = 0
         else:
@@ -937,7 +962,7 @@ class TestFtpFsOperations(unittest.TestCase):
         except ftplib.error_perm, err:
             self.failUnless("not retrievable" in str(err))
         else:
-            self.fail('Exception not raised')
+            self.fail('Exception not raised')       
 
 
 class TestFtpStoreData(unittest.TestCase):
@@ -986,9 +1011,11 @@ class TestFtpStoreData(unittest.TestCase):
         self.test_stor()
 
     def test_stor_ascii(self):
-        # Test STOR in ASCII mode.
+        # Test STOR in ASCII mode
 
         def store(cmd, fp, blocksize=8192):
+            # like storbinary() except it sends "type a" instead of
+            # "type i" before starting the transfer
             self.client.voidcmd('type a')
             conn = self.client.transfercmd(cmd)
             while 1:
@@ -1018,6 +1045,49 @@ class TestFtpStoreData(unittest.TestCase):
                 except (ftplib.Error, EOFError, socket.error):
                     safe_remove(TESTFN)
 
+    def test_stor_ascii_2(self):
+        # Test that no extra extra carriage returns are added to the
+        # file in ASCII mode in case CRLF gets truncated in two chunks
+        # (issue 116)
+
+        def store(cmd, fp, blocksize=8192):
+            # like storbinary() except it sends "type a" instead of
+            # "type i" before starting the transfer
+            self.client.voidcmd('type a')
+            conn = self.client.transfercmd(cmd)
+            while 1:
+                buf = fp.read(blocksize)
+                if not buf:
+                    break
+                conn.sendall(buf)
+            conn.close()
+            return self.client.voidresp()
+
+        old_buffer = ftpserver.DTPHandler.ac_in_buffer_size
+        try:
+            # set a small buffer so that CRLF gets delivered in two
+            # separate chunks: "CRLF", " f", "oo", " CR", "LF", " b", "ar"
+            ftpserver.DTPHandler.ac_in_buffer_size = 2
+            data = '\r\n foo \r\n bar'
+            self.dummy_sendfile.write(data)
+            self.dummy_sendfile.seek(0)
+            store('stor ' + TESTFN, self.dummy_sendfile)
+
+            expected = data.replace('\r\n', os.linesep)
+            self.client.retrbinary('retr ' + TESTFN, self.dummy_recvfile.write)
+            self.dummy_recvfile.seek(0)
+            self.assertEqual(expected, self.dummy_recvfile.read())
+        finally:
+            ftpserver.DTPHandler.ac_in_buffer_size = old_buffer
+            # We do not use os.remove() because file could still be
+            # locked by ftpd thread.  If DELE through FTP fails try
+            # os.remove() as last resort.
+            if os.path.exists(TESTFN):
+                try:
+                    self.client.delete(TESTFN)
+                except (ftplib.Error, EOFError, socket.error):
+                    safe_remove(TESTFN)
+
     def test_stou(self):
         data = 'abcde12345' * 100000
         self.dummy_sendfile.write(data)
@@ -1160,6 +1230,22 @@ class TestFtpStoreData(unittest.TestCase):
         self.dummy_sendfile.seek(0)
         self.client.storbinary('stor ' + TESTFN, self.dummy_sendfile)
 
+    def test_quit_during_transfer(self):
+        # RFC-959 states that if QUIT is sent while a transfer is in 
+        # progress, the connection must remain open for result response
+        # and the server will then close it.
+        conn = self.client.transfercmd('stor ' + TESTFN)
+        conn.sendall('abcde12345' * 50000)
+        self.client.sendcmd('quit')
+        conn.sendall('abcde12345' * 50000)
+        conn.close()
+        # expect the response (transfer ok)
+        self.client.voidresp()
+        # Make sure client has been disconnected.
+        # socket.error (Windows) or EOFError (Linux) exception is supposed
+        # to be raised in such a case.
+        self.assertRaises((socket.error, EOFError), self.client.sendcmd, 'noop')
+
 
 class TestFtpRetrieveData(unittest.TestCase):
     "Test RETR, REST, TYPE"
@@ -1191,6 +1277,11 @@ class TestFtpRetrieveData(unittest.TestCase):
         self.dummyfile.seek(0)
         self.assertEqual(hash(data), hash(self.dummyfile.read()))
 
+        # attempt to retrieve a file which doesn't exist
+        bogus = os.path.basename(tempfile.mktemp(dir=HOME))
+        self.assertRaises(ftplib.error_perm, self.client.retrbinary,
+                                             "retr " + bogus, lambda x: x)
+
     def test_retr_ascii(self):
         # Test RETR in ASCII mode.
 
@@ -1320,7 +1411,7 @@ class TestFtpListingCmds(unittest.TestCase):
         self.assertEqual(mlstline('mlst'), mlstline('mlst /'))
         # non-existent path
         bogus = os.path.basename(tempfile.mktemp(dir=HOME))
-        self.assertRaises(ftplib.error_perm, mlstline, bogus)
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'mlst '+bogus)
         # test file/dir notations
         self.failUnless('type=dir' in mlstline('mlst'))
         self.failUnless('type=file' in mlstline('mlst ' + TESTFN))
@@ -1349,6 +1440,24 @@ class TestFtpListingCmds(unittest.TestCase):
                 os.rmdir(dir)
             except OSError:
                 pass
+                
+    def test_mlsd_specific_platform_opts(self):
+        opts = []
+        feats = self.client.sendcmd('feat')
+        if 'unique' in feats:
+            opts.append('unique;')
+        if 'create' in feats:
+            opts.append('create;')
+        if 'unix.mode' in feats:
+            opts.append('unix.mode;')
+        if 'unix.uid' in feats:
+            opts.append('unix.uid;')
+        if 'unix.gid' in feats:
+            opts.append('unix.gid;')      
+        if opts:
+            lines = []
+            self.client.sendcmd('opts mlst ' + ''.join(opts))
+            self.client.retrlines('mlsd .', lines.append) 
 
     def test_stat(self):
         # Test STAT provided with argument which is equal to LIST
@@ -1360,6 +1469,25 @@ class TestFtpListingCmds(unittest.TestCase):
         bogus = os.path.basename(tempfile.mktemp(dir=HOME))
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stat ' + bogus)
 
+    def test_unforeseen_time_event(self):
+        # Emulate a case where the file last modification time is prior
+        # to year 1900.  This most likely will never happen unless
+        # someone specifically force the last modification time of a
+        # file in some way.
+        # To do so we temporarily override os.path.getmtime so that it
+        # returns a negative value referring to a year prior to 1900.
+        # It causes time.localtime/gmtime to raise a ValueError exception
+        # which is supposed to be handled by server.
+        _getmtime = ftpserver.AbstractedFS.getmtime
+        try:
+            ftpserver.AbstractedFS.getmtime = lambda x, y: -9000000000
+            self.client.sendcmd('stat /')  # test AbstractedFS.format_list()
+            self.client.sendcmd('mlst /')  # test AbstractedFS.format_mlsx()
+            # make sure client hasn't been disconnected
+            self.client.sendcmd('noop')
+        finally:
+            ftpserver.AbstractedFS.getmtime = _getmtime
+
 
 class TestFtpAbort(unittest.TestCase):
     "test: ABOR"
@@ -1445,6 +1573,80 @@ class TestFtpAbort(unittest.TestCase):
             self.assertEqual(self.client.getresp()[:3], '225')
 
 
+class ThrottleBandwidth(unittest.TestCase):
+    """Test ThrottledDTPHandler class."""
+
+    def setUp(self):
+        self.server = FTPd()
+        self.server.handler.dtp_handler = ftpserver.ThrottledDTPHandler
+        self.server.start()
+        self.client = ftplib.FTP()
+        self.client.connect(self.server.host, self.server.port)
+        self.client.sock.settimeout(2)
+        self.client.login(USER, PASSWD)
+
+        self.dummyfile = StringIO.StringIO()
+
+    def tearDown(self):
+        self.client.close()
+        self.server.handler.dtp_handler.read_limit = 0
+        self.server.handler.dtp_handler.write_limit = 0
+        self.server.handler.dtp_handler = ftpserver.DTPHandler
+        self.server.stop()
+
+        if not self.dummyfile.closed:
+            self.dummyfile.close()
+        if os.path.exists(TESTFN):
+            os.remove(TESTFN)
+
+    class CallNowInsteadOfLater(ftpserver.CallLater):
+        """A dirty hack I'm really ashamed of.
+        Force the original CallLater class to immediately "fire" any
+        callback, no matter what its actual timeout is.
+        This way we *don't* limit the transfer speed and run our test
+        faster, but for the underlying code implementing the bandwidth
+        throttling that doesn't make any difference.
+        """
+        def __le__(self, other):
+            self.timeout = 0
+            return 1
+
+    def test_throttle_send(self):
+        # This test doesn't test the actual speed accuracy, just
+        # awakes all that code which implements the throttling.
+        self.server.handler.dtp_handler.write_limit = 32768
+        data = 'abcde12345' * 100000
+        file = open(TESTFN, 'wb')
+        file.write(data)
+        file.close()
+
+        original_call_later = ftpserver.CallLater
+        ftpserver.CallLater = self.CallNowInsteadOfLater
+        try:
+            self.client.retrbinary("retr " + TESTFN, self.dummyfile.write)
+            self.dummyfile.seek(0)
+            self.assertEqual(hash(data), hash(self.dummyfile.read()))
+        finally:
+            ftpserver.CallLater = original_call_later
+
+    def test_throttle_recv(self):
+        # This test doesn't test the actual speed accuracy, just
+        # awakes all that code which implements the throttling.
+        self.server.handler.dtp_handler.read_limit = 32768
+        data = 'abcde12345' * 100000
+        self.dummyfile.write(data)
+        self.dummyfile.seek(0)
+
+        original_call_later = ftpserver.CallLater
+        ftpserver.CallLater = self.CallNowInsteadOfLater
+        try:
+            self.client.storbinary("stor " + TESTFN, self.dummyfile)
+            file_data = open(TESTFN, 'rb').read()
+            self.assertEqual(hash(data), hash(file_data))
+        finally:
+            ftpserver.CallLater = original_call_later
+
+
 class TestTimeouts(unittest.TestCase):
     """Test idle-timeout capabilities of control and data channels.
     Some tests may fail on slow machines.
@@ -1498,6 +1700,22 @@ class TestTimeouts(unittest.TestCase):
         # ensure client has been kicked off
         self.assertRaises((socket.error, EOFError), self.client.sendcmd, 'noop')
 
+    def test_data_timeout_not_reached(self):
+        # Impose a timeout for the data channel, then keep sending data for a 
+        # time which is longer than that to make sure that the code checking
+        # whether the transfer stalled for with no progress is executed.
+        self._setUp(data_timeout=0.1)
+        sock = self.client.transfercmd('stor ' + TESTFN)
+        try:
+            stop_at = time.time() + 0.2
+            while time.time() < stop_at:
+                sock.send('x' * 1024)
+            sock.close()
+            self.client.voidresp()
+        finally:
+            if os.path.exists(TESTFN):
+                self.client.delete(TESTFN)
+
     def test_idle_data_timeout1(self):
         # Tests that the control connection timeout is suspended while
         # the data channel is opened
@@ -1581,6 +1799,7 @@ class TestConfigurableOptions(unittest.TestCase):
     def tearDown(self):
         # set back options to their original value
         self.server.server.max_cons = 0
+        self.server.server.max_cons_per_ip = 0
         self.server.handler.banner = "pyftpdlib %s ready." %ftpserver.__ver__
         self.server.handler.max_login_attempts = 3
         self.server.handler._auth_failed_timeout = 5
@@ -1621,6 +1840,30 @@ class TestConfigurableOptions(unittest.TestCase):
             c1.close()
             c2.close()
             c3.close()
+            
+    def test_max_connections_per_ip(self):
+        # Test FTPServer.max_cons_per_ip attribute
+        self.server.server.max_cons_per_ip = 3
+        self.client.quit()
+        c1 = ftplib.FTP()
+        c2 = ftplib.FTP()
+        c3 = ftplib.FTP()
+        c4 = ftplib.FTP()
+        try:
+            c1.connect(self.server.host, self.server.port)
+            c2.connect(self.server.host, self.server.port)
+            c3.connect(self.server.host, self.server.port)
+            self.assertRaises(ftplib.error_temp, c4.connect, self.server.host,
+                              self.server.port)
+            # Make sure client has been disconnected.
+            # socket.error (Windows) or EOFError (Linux) exception is
+            # supposed to be raised in such a case.
+            self.assertRaises((socket.error, EOFError), c4.sendcmd, 'noop')
+        finally:
+            c1.close()
+            c2.close()
+            c3.close()
+            c4.close()
 
     def test_banner(self):
         # Test FTPHandler.banner attribute
@@ -1658,6 +1901,16 @@ class TestConfigurableOptions(unittest.TestCase):
         self.assert_(self.client.makepasv()[1] in _range)
         self.assert_(self.client.makepasv()[1] in _range)
 
+    def test_passive_ports_busy(self):
+        # If the ports in the configured range are busy it is expected
+        # that a kernel-assigned port gets chosen
+        s = socket.socket()
+        s.bind((HOST, 0))
+        port = s.getsockname()[1]
+        self.server.handler.passive_ports = [port]
+        resulting_port = self.client.makepasv()[1]
+        self.assert_(port != resulting_port)
+
     if hasattr(socket, 'getservbyport'):   # python > 2.3
 
         def test_permit_privileged_ports(self):
@@ -1704,6 +1957,72 @@ class TestConfigurableOptions(unittest.TestCase):
                     sock.close()
 
 
+class TestCallbacks(unittest.TestCase):
+    """Test FTPHandler class callback methods."""
+
+    def _setUp(self, handler):
+        self.server = FTPd(handler=handler)
+        self.server.start()
+        self.client = ftplib.FTP()
+        self.client.connect(self.server.host, self.server.port)
+        self.client.sock.settimeout(2)
+        self.client.login(USER, PASSWD)
+        self.file = open(TESTFN, 'w+b')
+        self.dummyfile = StringIO.StringIO()
+        self._tearDown = False
+
+    def tearDown(self):
+        if not self._tearDown:
+            self._tearDown = True
+            if self.client is not None:
+                self.client.close()
+            if self.server is not None:
+                self.server.stop()
+            if not self.file.closed:
+                self.file.close()
+            if not self.dummyfile.closed:
+                self.dummyfile.close()
+            os.remove(TESTFN)
+
+    def test_on_file_sent(self):
+        _file = []
+
+        class TestHandler(ftpserver.FTPHandler):
+            def on_file_sent(self, file):
+                _file.append(file)
+
+            def on_file_received(self, file):
+                raise Exception
+
+        self._setUp(TestHandler)
+        data = 'abcde12345' * 100000
+        self.file.write(data)
+        self.file.close()
+        self.client.retrbinary("retr " + TESTFN, lambda x: x)
+        # shut down the server to avoid race conditions
+        self.tearDown()
+        self.assertEqual(_file.pop(), os.path.abspath(TESTFN))
+
+    def test_on_file_received(self):
+        _file = []
+
+        class TestHandler(ftpserver.FTPHandler):
+            def on_file_sent(self, file):
+                raise Exception
+
+            def on_file_received(self, file):
+                _file.append(file)
+
+        self._setUp(TestHandler)
+        data = 'abcde12345' * 100000
+        self.dummyfile.write(data)
+        self.dummyfile.seek(0)
+        self.client.storbinary('stor ' + TESTFN, self.dummyfile)
+        # shut down the server to avoid race conditions
+        self.tearDown()
+        self.assertEqual(_file.pop(), os.path.abspath(TESTFN))
+
+
 class _TestNetworkProtocols(unittest.TestCase):
     """Test PASV, EPSV, PORT and EPRT commands.
 
@@ -1762,6 +2081,19 @@ class _TestNetworkProtocols(unittest.TestCase):
         # port < 1024
         self.assertEqual(self.cmdresp('eprt |%s|%s|222|' %(self.proto,
                        self.HOST)), "501 Can't connect over a privileged port.")
+        # proto > 2
+        _cmd = 'eprt |3|%s|%s|' %(self.server.host, self.server.port)
+        self.assertRaises(ftplib.error_perm,  self.client.sendcmd, _cmd)
+
+
+        if self.proto == '1':
+            # len(ip.octs) > 4
+            self.assertEqual(self.cmdresp('eprt |1|1.2.3.4.5|2048|'), msg)
+            # ip.oct > 255
+            self.assertEqual(self.cmdresp('eprt |1|1.2.3.256|2048|'), msg)
+            # bad proto
+            resp = self.cmdresp('eprt |2|1.2.3.256|2048|')
+            self.assert_("Network protocol not supported" in resp)
 
         # test connection
         sock = socket.socket(self.client.af, socket.SOCK_STREAM)
@@ -1787,6 +2119,9 @@ class _TestNetworkProtocols(unittest.TestCase):
         else:
             self.fail("Exception not raised")
 
+        # proto > 2
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'epsv 3')
+
         # test connection
         for cmd in ('EPSV', 'EPSV ' + self.proto):
             host, port = ftplib.parse229(self.client.sendcmd(cmd),
@@ -1881,7 +2216,8 @@ class FTPd(threading.Thread):
     eventually re-start() the server.
     """
 
-    def __init__(self, host=HOST, port=0, verbose=False):
+    def __init__(self, host=HOST, port=0, handler=ftpserver.FTPHandler,
+                 verbose=False):
         threading.Thread.__init__(self)
         self.__serving = False
         self.__stopped = False
@@ -1893,7 +2229,7 @@ class FTPd(threading.Thread):
         self.authorizer = ftpserver.DummyAuthorizer()
         self.authorizer.add_user(USER, PASSWD, HOME, perm='elradfmw')  # full perms
         self.authorizer.add_anonymous(HOME)
-        self.handler = ftpserver.FTPHandler
+        self.handler = handler
         self.handler.authorizer = self.authorizer
         self.server = ftpserver.FTPServer((host, port), self.handler)
         self.host, self.port = self.server.socket.getsockname()[:2]
@@ -1958,9 +2294,11 @@ def test_main(tests=None):
                  TestFtpStoreData,
                  TestFtpRetrieveData,
                  TestFtpListingCmds,
+                 ThrottleBandwidth,
                  TestFtpAbort,
                  TestTimeouts,
-                 TestConfigurableOptions
+                 TestConfigurableOptions,
+                 TestCallbacks
                  ]
         if SUPPORTS_IPV4:
             tests.append(TestIPv4Environment)
